@@ -37,11 +37,6 @@
  * compositor needs to delay hiding the windows until the switch
  * workspace animation completes.
  *
- * meta_compositor_maximize_window() and meta_compositor_unmaximize_window()
- * are transitions within the visible state. The window is resized __before__
- * the call, so it may be necessary to readjust the display based on the
- * old_rect to start the animation.
- *
  * # Containers #
  *
  * There's two containers in the stage that are used to place window actors, here
@@ -79,7 +74,6 @@
 #include "frame.h"
 #include <X11/extensions/shape.h>
 #include <X11/extensions/Xcomposite.h>
-#include "meta-sync-ring.h"
 
 #include "backends/x11/meta-backend-x11.h"
 
@@ -126,11 +120,7 @@ meta_switch_workspace_completed (MetaCompositor *compositor)
 void
 meta_compositor_destroy (MetaCompositor *compositor)
 {
-  clutter_threads_remove_repaint_func (compositor->pre_paint_func_id);
-  clutter_threads_remove_repaint_func (compositor->post_paint_func_id);
-
-  if (compositor->have_x11_sync_object)
-    meta_sync_ring_destroy ();
+  clutter_threads_remove_repaint_func (compositor->repaint_func_id);
 }
 
 static void
@@ -473,11 +463,13 @@ meta_compositor_manage (MetaCompositor *compositor)
   MetaDisplay *display = compositor->display;
   Display *xdisplay = display->xdisplay;
   MetaScreen *screen = display->screen;
-  MetaBackend *backend = meta_get_backend ();
 
   meta_screen_set_cm_selection (display->screen);
 
-  compositor->stage = meta_backend_get_stage (backend);
+  {
+    MetaBackend *backend = meta_get_backend ();
+    compositor->stage = meta_backend_get_stage (backend);
+  }
 
   /* We use connect_after() here to accomodate code in GNOME Shell that,
    * when benchmarking drawing performance, connects to ::after-paint
@@ -513,7 +505,7 @@ meta_compositor_manage (MetaCompositor *compositor)
 
       compositor->output = screen->composite_overlay_window;
 
-      xwin = meta_backend_x11_get_xwindow (META_BACKEND_X11 (backend));
+      xwin = meta_backend_x11_get_xwindow (META_BACKEND_X11 (meta_get_backend ()));
 
       XReparentWindow (xdisplay, xwin, compositor->output, 0, 0);
 
@@ -533,8 +525,6 @@ meta_compositor_manage (MetaCompositor *compositor)
        * contents until we show the stage.
        */
       XMapWindow (xdisplay, compositor->output);
-
-      compositor->have_x11_sync_object = meta_sync_ring_init (xdisplay);
     }
 
   redirect_windows (display->screen);
@@ -736,9 +726,6 @@ meta_compositor_process_event (MetaCompositor *compositor,
         process_damage (compositor, (XDamageNotifyEvent *) event, window);
     }
 
-  if (compositor->have_x11_sync_object)
-    meta_sync_ring_handle_event (event);
-
   /* Clutter needs to know about MapNotify events otherwise it will
      think the stage is invisible */
   if (!meta_is_wayland_compositor () && event->type == MapNotify)
@@ -777,23 +764,14 @@ meta_compositor_hide_window (MetaCompositor *compositor,
 }
 
 void
-meta_compositor_maximize_window (MetaCompositor    *compositor,
-                                 MetaWindow        *window,
-				 MetaRectangle	   *old_rect,
-				 MetaRectangle	   *new_rect)
+meta_compositor_size_change_window (MetaCompositor    *compositor,
+                                    MetaWindow        *window,
+                                    MetaSizeChange     which_change,
+                                    MetaRectangle     *old_frame_rect,
+                                    MetaRectangle     *old_buffer_rect)
 {
   MetaWindowActor *window_actor = META_WINDOW_ACTOR (meta_window_get_compositor_private (window));
-  meta_window_actor_maximize (window_actor, old_rect, new_rect);
-}
-
-void
-meta_compositor_unmaximize_window (MetaCompositor    *compositor,
-                                   MetaWindow        *window,
-				   MetaRectangle     *old_rect,
-				   MetaRectangle     *new_rect)
-{
-  MetaWindowActor *window_actor = META_WINDOW_ACTOR (meta_window_get_compositor_private (window));
-  meta_window_actor_unmaximize (window_actor, old_rect, new_rect);
+  meta_window_actor_size_change (window_actor, META_SIZE_CHANGE_MAXIMIZE, old_frame_rect, old_buffer_rect);
 }
 
 void
@@ -1052,12 +1030,11 @@ frame_callback (CoglOnscreen  *onscreen,
     }
 }
 
-static gboolean
-meta_pre_paint_func (gpointer data)
+static void
+pre_paint_windows (MetaCompositor *compositor)
 {
   GList *l;
   MetaWindowActor *top_window;
-  MetaCompositor *compositor = data;
 
   if (compositor->onscreen == NULL)
     {
@@ -1069,7 +1046,7 @@ meta_pre_paint_func (gpointer data)
     }
 
   if (compositor->windows == NULL)
-    return TRUE;
+    return;
 
   top_window = g_list_last (compositor->windows)->data;
 
@@ -1086,12 +1063,10 @@ meta_pre_paint_func (gpointer data)
     {
       /* We need to make sure that any X drawing that happens before
        * the XDamageSubtract() for each window above is visible to
-       * subsequent GL rendering; the standardized way to do this is
-       * GL_EXT_X11_sync_object. Since this isn't implemented yet in
-       * mesa, we also have a path that relies on the implementation
-       * of the open source drivers.
-       *
-       * Anything else, we just hope for the best.
+       * subsequent GL rendering; the only standardized way to do this
+       * is EXT_x11_sync_object, which isn't yet widely available. For
+       * now, we count on details of Xorg and the open source drivers,
+       * and hope for the best otherwise.
        *
        * Xorg and open source driver specifics:
        *
@@ -1106,28 +1081,17 @@ meta_pre_paint_func (gpointer data)
        * round trip request at this point is sufficient to flush the
        * GLX buffers.
        */
-      if (compositor->have_x11_sync_object)
-        compositor->have_x11_sync_object = meta_sync_ring_insert_wait ();
-      else
-        XSync (compositor->display->xdisplay, False);
-    }
-
-  return TRUE;
-}
-
-static gboolean
-meta_post_paint_func (gpointer data)
-{
-  MetaCompositor *compositor = data;
-
-  if (compositor->frame_has_updated_xsurfaces)
-    {
-      if (compositor->have_x11_sync_object)
-        compositor->have_x11_sync_object = meta_sync_ring_after_frame ();
+      XSync (compositor->display->xdisplay, False);
 
       compositor->frame_has_updated_xsurfaces = FALSE;
     }
+}
 
+static gboolean
+meta_repaint_func (gpointer data)
+{
+  MetaCompositor *compositor = data;
+  pre_paint_windows (compositor);
   return TRUE;
 }
 
@@ -1162,16 +1126,10 @@ meta_compositor_new (MetaDisplay *display)
                     G_CALLBACK (on_shadow_factory_changed),
                     compositor);
 
-  compositor->pre_paint_func_id =
-    clutter_threads_add_repaint_func_full (CLUTTER_REPAINT_FLAGS_PRE_PAINT,
-                                           meta_pre_paint_func,
-                                           compositor,
-                                           NULL);
-  compositor->post_paint_func_id =
-    clutter_threads_add_repaint_func_full (CLUTTER_REPAINT_FLAGS_POST_PAINT,
-                                           meta_post_paint_func,
-                                           compositor,
-                                           NULL);
+  compositor->repaint_func_id = clutter_threads_add_repaint_func (meta_repaint_func,
+                                                                  compositor,
+                                                                  NULL);
+
   return compositor;
 }
 
