@@ -40,7 +40,7 @@
 #include "keybindings-private.h"
 #include "stack.h"
 #include <meta/compositor.h>
-#include "mutter-enum-types.h"
+#include <meta/meta-enum-types.h>
 #include "core.h"
 #include "meta-cursor-tracker-private.h"
 
@@ -1439,8 +1439,8 @@ meta_screen_get_monitor_for_rect (MetaScreen    *screen,
 }
 
 const MetaMonitorInfo*
-meta_screen_get_monitor_for_window (MetaScreen *screen,
-                                    MetaWindow *window)
+meta_screen_calculate_monitor_for_window (MetaScreen *screen,
+                                          MetaWindow *window)
 {
   MetaRectangle window_rect;
 
@@ -2388,12 +2388,12 @@ on_monitors_changed (MetaMonitorManager *manager,
                        &changes);
     }
 
-  /* Queue a resize on all the windows */
-  meta_screen_foreach_window (screen, META_LIST_DEFAULT, meta_screen_resize_func, 0);
-
   /* Fix up monitor for all windows on this screen */
   meta_screen_foreach_window (screen, META_LIST_INCLUDE_OVERRIDE_REDIRECT, (MetaScreenWindowFunc) meta_window_update_for_monitors_changed, 0);
 
+  /* Queue a resize on all the windows */
+  meta_screen_foreach_window (screen, META_LIST_DEFAULT, meta_screen_resize_func, 0);
+
   meta_screen_queue_check_fullscreen (screen);
 
   g_signal_emit (screen, screen_signals[MONITORS_CHANGED], 0);
