@@ -78,6 +78,7 @@ typedef enum
   META_MOVE_RESIZE_RESIZE_ACTION     = 1 << 3,
   META_MOVE_RESIZE_WAYLAND_RESIZE    = 1 << 4,
   META_MOVE_RESIZE_STATE_CHANGED     = 1 << 5,
+  META_MOVE_RESIZE_DONT_SYNC_COMPOSITOR = 1 << 6,
 } MetaMoveResizeFlags;
 
 typedef enum
@@ -481,6 +482,7 @@ struct _MetaWindowClass
   gboolean (*update_icon)        (MetaWindow       *window,
                                   cairo_surface_t **icon,
                                   cairo_surface_t **mini_icon);
+  void (*update_main_monitor)    (MetaWindow *window);
   void (*main_monitor_changed)   (MetaWindow *window,
                                   const MetaMonitorInfo *old);
 };
@@ -694,4 +696,6 @@ void meta_window_set_alive (MetaWindow *window, gboolean is_alive);
 
 gboolean meta_window_has_pointer (MetaWindow *window);
 
+void meta_window_emit_size_changed (MetaWindow *window);
+
 #endif
