@@ -42,6 +42,7 @@
 #include "meta-wayland-pointer.h"
 #include "meta-wayland-popup.h"
 #include "meta-wayland-data-device.h"
+#include "meta-wayland-outputs.h"
 
 #include "meta-cursor-tracker-private.h"
 #include "display-private.h"
@@ -187,32 +188,35 @@ calculate_surface_window_geometry (MetaWaylandSurface *surface,
                                    float               parent_x,
                                    float               parent_y)
 {
-  ClutterActor *surface_actor = CLUTTER_ACTOR (surface->surface_actor);
+  MetaSurfaceActorWayland *surface_actor =
+    META_SURFACE_ACTOR_WAYLAND (surface->surface_actor);
+  MetaRectangle subsurface_rect;
   MetaRectangle geom;
-  float x, y;
   GList *l;
 
   /* Unmapped surfaces don't count. */
-  if (!CLUTTER_ACTOR_IS_VISIBLE (surface_actor))
+  if (!CLUTTER_ACTOR_IS_VISIBLE (CLUTTER_ACTOR (surface_actor)))
     return;
 
   if (!surface->buffer)
     return;
 
-  /* XXX: Is there a better way to do this using Clutter APIs? */
-  clutter_actor_get_position (surface_actor, &x, &y);
+  meta_surface_actor_wayland_get_subsurface_rect (surface_actor,
+                                                  &subsurface_rect);
 
-  geom.x = parent_x + x;
-  geom.y = parent_x + y;
-  geom.width = cogl_texture_get_width (surface->buffer->texture);
-  geom.height = cogl_texture_get_height (surface->buffer->texture);
+  geom.x = parent_x + subsurface_rect.x;
+  geom.y = parent_x + subsurface_rect.y;
+  geom.width = subsurface_rect.width;
+  geom.height = subsurface_rect.height;
 
   meta_rectangle_union (total_geometry, &geom, total_geometry);
 
   for (l = surface->subsurfaces; l != NULL; l = l->next)
     {
       MetaWaylandSurface *subsurface = l->data;
-      calculate_surface_window_geometry (subsurface, total_geometry, x, y);
+      calculate_surface_window_geometry (subsurface, total_geometry,
+                                         subsurface_rect.x,
+                                         subsurface_rect.y);
     }
 }
 
@@ -381,18 +385,13 @@ static void
 subsurface_surface_commit (MetaWaylandSurface      *surface,
                            MetaWaylandPendingState *pending)
 {
-  MetaSurfaceActor *surface_actor = surface->surface_actor;
-  float x, y;
+  MetaSurfaceActorWayland *surface_actor =
+    META_SURFACE_ACTOR_WAYLAND (surface->surface_actor);
 
   if (surface->buffer != NULL)
     clutter_actor_show (CLUTTER_ACTOR (surface_actor));
   else
     clutter_actor_hide (CLUTTER_ACTOR (surface_actor));
-
-  clutter_actor_get_position (CLUTTER_ACTOR (surface_actor), &x, &y);
-  x += pending->dx;
-  y += pending->dy;
-  clutter_actor_set_position (CLUTTER_ACTOR (surface_actor), x, y);
 }
 
 /* A non-subsurface is always desynchronized.
@@ -426,20 +425,23 @@ parent_surface_state_applied (gpointer data, gpointer user_data)
 
   if (surface->sub.pending_pos)
     {
-      clutter_actor_set_position (CLUTTER_ACTOR (surface->surface_actor),
-                                  surface->sub.pending_x,
-                                  surface->sub.pending_y);
+      surface->sub.x = surface->sub.pending_x;
+      surface->sub.y = surface->sub.pending_y;
       surface->sub.pending_pos = FALSE;
     }
 
   if (surface->sub.pending_placement_ops)
     {
       GSList *it;
+      MetaWaylandSurface *parent = surface->sub.parent;
+      ClutterActor *parent_actor =
+        clutter_actor_get_parent (CLUTTER_ACTOR (parent->surface_actor));
+      ClutterActor *surface_actor =
+        surface_actor = CLUTTER_ACTOR (surface->surface_actor);
+
       for (it = surface->sub.pending_placement_ops; it; it = it->next)
         {
           MetaWaylandSubsurfacePlacementOp *op = it->data;
-          ClutterActor *surface_actor;
-          ClutterActor *parent_actor;
           ClutterActor *sibling_actor;
 
           if (!op->sibling)
@@ -448,8 +450,6 @@ parent_surface_state_applied (gpointer data, gpointer user_data)
               continue;
             }
 
-          surface_actor = CLUTTER_ACTOR (surface->surface_actor);
-          parent_actor = clutter_actor_get_parent (CLUTTER_ACTOR (surface->sub.parent));
           sibling_actor = CLUTTER_ACTOR (op->sibling->surface_actor);
 
           switch (op->placement)
@@ -476,6 +476,9 @@ parent_surface_state_applied (gpointer data, gpointer user_data)
 
   if (is_surface_effectively_synchronized (surface))
     apply_pending_state (surface, &surface->sub.pending);
+
+  meta_surface_actor_wayland_sync_subsurface_state (
+    META_SURFACE_ACTOR_WAYLAND (surface->surface_actor));
 }
 
 static void
@@ -521,21 +524,32 @@ apply_pending_state (MetaWaylandSurface      *surface,
       surface->input_region = cairo_region_reference (pending->input_region);
     }
 
-  meta_surface_actor_wayland_sync_state (
-    META_SURFACE_ACTOR_WAYLAND (surface->surface_actor));
-
   /* wl_surface.frame */
   wl_list_insert_list (&compositor->frame_callbacks, &pending->frame_callback_list);
   wl_list_init (&pending->frame_callback_list);
 
-  if (surface == compositor->seat->pointer.cursor_surface)
-    cursor_surface_commit (surface, pending);
-  else if (meta_wayland_data_device_is_dnd_surface (&compositor->seat->data_device, surface))
-    dnd_surface_commit (surface, pending);
-  else if (surface->window)
-    toplevel_surface_commit (surface, pending);
-  else if (surface->wl_subsurface)
-    subsurface_surface_commit (surface, pending);
+  switch (surface->role)
+    {
+    case META_WAYLAND_SURFACE_ROLE_NONE:
+      break;
+    case META_WAYLAND_SURFACE_ROLE_CURSOR:
+      cursor_surface_commit (surface, pending);
+      break;
+    case META_WAYLAND_SURFACE_ROLE_DND:
+      dnd_surface_commit (surface, pending);
+      break;
+    case META_WAYLAND_SURFACE_ROLE_XDG_SURFACE:
+    case META_WAYLAND_SURFACE_ROLE_XDG_POPUP:
+    case META_WAYLAND_SURFACE_ROLE_WL_SHELL_SURFACE:
+      toplevel_surface_commit (surface, pending);
+      break;
+    case META_WAYLAND_SURFACE_ROLE_SUBSURFACE:
+      subsurface_surface_commit (surface, pending);
+      break;
+    }
+
+  meta_surface_actor_wayland_sync_state (
+    META_SURFACE_ACTOR_WAYLAND (surface->surface_actor));
 
   pending_state_reset (pending);
 
@@ -762,6 +776,114 @@ sync_drag_dest_funcs (MetaWaylandSurface *surface)
     surface->dnd.funcs = meta_wayland_data_device_get_drag_dest_funcs ();
 }
 
+static void
+surface_entered_output (MetaWaylandSurface *surface,
+                        MetaWaylandOutput *wayland_output)
+{
+  GList *iter;
+  struct wl_resource *resource;
+
+  for (iter = wayland_output->resources; iter != NULL; iter = iter->next)
+    {
+      resource = iter->data;
+
+      if (wl_resource_get_client (resource) !=
+          wl_resource_get_client (surface->resource))
+        continue;
+
+      wl_surface_send_enter (surface->resource, resource);
+    }
+}
+
+static void
+surface_left_output (MetaWaylandSurface *surface,
+                     MetaWaylandOutput *wayland_output)
+{
+  GList *iter;
+  struct wl_resource *resource;
+
+  for (iter = wayland_output->resources; iter != NULL; iter = iter->next)
+    {
+      resource = iter->data;
+
+      if (wl_resource_get_client (resource) !=
+          wl_resource_get_client (surface->resource))
+        continue;
+
+      wl_surface_send_leave (surface->resource, resource);
+    }
+}
+
+static void
+set_surface_is_on_output (MetaWaylandSurface *surface,
+                          MetaWaylandOutput *wayland_output,
+                          gboolean is_on_output);
+
+static void
+surface_handle_output_destroy (MetaWaylandOutput *wayland_output,
+                               GParamSpec *pspec,
+                               MetaWaylandSurface *surface)
+{
+  set_surface_is_on_output (surface, wayland_output, FALSE);
+}
+
+static void
+set_surface_is_on_output (MetaWaylandSurface *surface,
+                          MetaWaylandOutput *wayland_output,
+                          gboolean is_on_output)
+{
+  gboolean was_on_output = g_hash_table_contains (surface->outputs,
+                                                  wayland_output);
+
+  if (!was_on_output && is_on_output)
+    {
+      g_signal_connect (wayland_output, "output-destroyed",
+                        G_CALLBACK (surface_handle_output_destroy),
+                        surface);
+      g_hash_table_add (surface->outputs, wayland_output);
+      surface_entered_output (surface, wayland_output);
+    }
+  else if (was_on_output && !is_on_output)
+    {
+      g_hash_table_remove (surface->outputs, wayland_output);
+      g_signal_handlers_disconnect_by_func  (
+        wayland_output, (gpointer)surface_handle_output_destroy, surface);
+      surface_left_output (surface, wayland_output);
+    }
+}
+
+static void
+update_surface_output_state (gpointer key, gpointer value, gpointer user_data)
+{
+  MetaWaylandOutput *wayland_output = value;
+  MetaWaylandSurface *surface = user_data;
+  MetaSurfaceActorWayland *actor =
+    META_SURFACE_ACTOR_WAYLAND (surface->surface_actor);
+  MetaMonitorInfo *monitor;
+  gboolean is_on_output;
+
+  monitor = wayland_output->monitor_info;
+  if (!monitor)
+    {
+      set_surface_is_on_output (surface, wayland_output, FALSE);
+      return;
+    }
+
+  is_on_output = meta_surface_actor_wayland_is_on_monitor (actor, monitor);
+  set_surface_is_on_output (surface, wayland_output, is_on_output);
+}
+
+void
+meta_wayland_surface_update_outputs (MetaWaylandSurface *surface)
+{
+  if (!surface->compositor)
+    return;
+
+  g_hash_table_foreach (surface->compositor->outputs,
+                        update_surface_output_state,
+                        surface);
+}
+
 void
 meta_wayland_surface_set_window (MetaWaylandSurface *surface,
                                  MetaWindow         *window)
@@ -791,10 +913,15 @@ wl_surface_destructor (struct wl_resource *resource)
   if (surface->input_region)
     cairo_region_destroy (surface->input_region);
 
+  meta_surface_actor_wayland_surface_destroyed (
+    META_SURFACE_ACTOR_WAYLAND (surface->surface_actor));
+
   g_object_unref (surface->surface_actor);
 
   meta_wayland_compositor_destroy_frame_callbacks (compositor, surface);
 
+  g_hash_table_unref (surface->outputs);
+
   if (surface->resource)
     wl_resource_set_user_data (surface->resource, NULL);
 
@@ -833,6 +960,8 @@ meta_wayland_surface_create (MetaWaylandCompositor *compositor,
 
   sync_drag_dest_funcs (surface);
 
+  surface->outputs = g_hash_table_new (NULL, NULL);
+
   pending_state_init (&surface->pending);
   return surface;
 }
@@ -1275,11 +1404,8 @@ xdg_shell_get_xdg_popup (struct wl_client *client,
                                     &surface->popup.parent_destroy_listener);
 
   window = meta_window_wayland_new (display, surface);
-  meta_window_move_frame (window, FALSE,
-                          parent_surf->window->buffer_rect.x + x,
-                          parent_surf->window->buffer_rect.y + y);
+  meta_window_wayland_place_relative_to (window, parent_surf->window, x, y);
   window->showing_for_first_time = FALSE;
-  window->placed = TRUE;
 
   meta_wayland_surface_set_window (surface, window);
 
@@ -1443,10 +1569,9 @@ wl_shell_surface_set_transient (struct wl_client *client,
   wl_shell_surface_set_state (surface, SURFACE_STATE_TOPLEVEL);
 
   meta_window_set_transient_for (surface->window, parent_surf->window);
-  meta_window_move_frame (surface->window, FALSE,
-                          parent_surf->window->rect.x + x,
-                          parent_surf->window->rect.y + y);
-  surface->window->placed = TRUE;
+  meta_window_wayland_place_relative_to (surface->window,
+                                         parent_surf->window,
+                                         x, y);
 }
 
 static void
@@ -1461,6 +1586,17 @@ wl_shell_surface_set_fullscreen (struct wl_client *client,
   wl_shell_surface_set_state (surface, SURFACE_STATE_FULLSCREEN);
 }
 
+static void
+handle_wl_shell_popup_parent_destroyed (struct wl_listener *listener,
+                                        void *data)
+{
+  MetaWaylandSurface *surface =
+    wl_container_of (listener, surface, popup.parent_destroy_listener);
+
+  wl_list_remove (&surface->popup.parent_destroy_listener.link);
+  surface->popup.parent = NULL;
+}
+
 static void
 wl_shell_surface_set_popup (struct wl_client *client,
                             struct wl_resource *resource,
@@ -1484,10 +1620,18 @@ wl_shell_surface_set_popup (struct wl_client *client,
     }
 
   meta_window_set_transient_for (surface->window, parent_surf->window);
-  meta_window_move_frame (surface->window, FALSE,
-                          parent_surf->window->rect.x + x,
-                          parent_surf->window->rect.y + y);
-  surface->window->placed = TRUE;
+  meta_window_wayland_place_relative_to (surface->window,
+                                         parent_surf->window,
+                                         x, y);
+
+  if (!surface->popup.parent)
+    {
+      surface->popup.parent = parent_surf;
+      surface->popup.parent_destroy_listener.notify =
+        handle_wl_shell_popup_parent_destroyed;
+      wl_resource_add_destroy_listener (parent_surf->resource,
+                                        &surface->popup.parent_destroy_listener);
+    }
 
   meta_wayland_pointer_start_popup_grab (&seat->pointer, surface);
 }
@@ -2011,12 +2155,6 @@ meta_wayland_surface_configure_notify (MetaWaylandSurface *surface,
       wl_array_init (&states);
       fill_states (&states, surface->window);
 
-      /* new_width and new_height comes from window->rect, which is based on
-       * the buffer size, not the surface size. The configure event requires
-       * surface size. */
-      new_width /= surface->scale;
-      new_height /= surface->scale;
-
       xdg_surface_send_configure (surface->xdg_surface, new_width, new_height, &states, serial);
 
       wl_array_release (&states);
@@ -2102,3 +2240,22 @@ meta_wayland_surface_drag_dest_drop (MetaWaylandSurface *surface)
 
   surface->dnd.funcs->drop (data_device, surface);
 }
+
+MetaWindow *
+meta_wayland_surface_get_toplevel_window (MetaWaylandSurface *surface)
+{
+  while (surface)
+    {
+      if (surface->window)
+        {
+          if (surface->popup.parent)
+            surface = surface->popup.parent;
+          else
+            return surface->window;
+        }
+      else
+        surface = surface->sub.parent;
+    }
+
+  return NULL;
+}
