@@ -100,6 +100,7 @@ struct _MetaWaylandSurface
   int scale;
   int32_t offset_x, offset_y;
   GList *subsurfaces;
+  GHashTable *outputs;
 
   struct {
     const MetaWaylandDragDestFuncs *funcs;
@@ -135,6 +136,9 @@ struct _MetaWaylandSurface
     MetaWaylandSurface *parent;
     struct wl_listener parent_destroy_listener;
 
+    int x;
+    int y;
+
     /* When the surface is synchronous, its state will be applied
      * when the parent is committed. This is done by moving the
      * "real" pending state below to here when this surface is
@@ -187,4 +191,8 @@ void                meta_wayland_surface_drag_dest_motion    (MetaWaylandSurface
 void                meta_wayland_surface_drag_dest_focus_out (MetaWaylandSurface   *surface);
 void                meta_wayland_surface_drag_dest_drop      (MetaWaylandSurface   *surface);
 
+void                meta_wayland_surface_update_outputs (MetaWaylandSurface *surface);
+
+MetaWindow *        meta_wayland_surface_get_toplevel_window (MetaWaylandSurface *surface);
+
 #endif
